function varargout = exchange(Sys,Exp,Opt)

% Implements intramolecular chemical exchange using theory from
%   Binsch, J.Am.Chem.Soc. 91, 1304 (1969)
%   Heinzer, Mol.Phys. 22, 167 (1971)
%   Bain, Prog. NMR Spectr. 43, 63 (2003)

% Methods:
% -3 ?
% -2 ?
% -1 Liouville method, two-site mutual exchange, Matlab accumulation
%  0 Liouville method, two-site mutual exchange, mex accumulation
%  1 Liouville method, general n-site exchange, mex accumulation
%  2 Liouville method, general exchange, Matlab accumulation
%  3 Bloch/McConnell, two-site mutual exchange

% Stefan Stoll, 2014

if (nargin<2), Exp = struct; end
if (nargin<3), Opt = struct; end
if ~isstruct(Sys), error('Spin system structure Sys is missing.'); end
if ~isstruct(Exp), error('Experiment structure Exp is missing.'); end

% Spin system
%------------------------------------------------------------
if ~isfield(Sys,'g'), Sys.g = gfree; end
if ~isfield(Sys,'Nucs'), Sys.Nucs = ''; end

if ~isempty(Sys.Nucs)
  I = nucspin(Sys.Nucs);
  if ~isfield(Sys,'n'), Sys.n = ones(size(I)); end
  nNucGroups = size(Sys.A,2);
  if numel(I)~=nNucGroups
    error('Sys.A and Sys.Nucs are not consistent.');
  end
  if isfield(Sys,'xc')
    if size(Sys.xc,2)~=nNucGroups
      error('Sys.xc and Sys.A must have the same number of elements.');
    end
    for iSite = 1:size(Sys.xc,1)
      p = Sys.xc(iSite,:); % permutation
      if any(sort(p)~=1:nNucGroups)
        error('Row %d of Sys.xc is incorrect: this is not a permutation of the nuclei.',iSite);
      end
      if any(Sys.n~=Sys.n(p))
        error('Row %d of Sys.xc is incorrect: nuclei have different multiplicity and cannot be swapped.',iSite);
      end
      if any(I~=I(p))
        error('Row %d of Sys.xc is incorrect: nuclei are different and cannot be swapped',iSite);
      end
    end
    xc = [1:nNucGroups; Sys.xc]; % supplement identity permutation if absent
    xc = unique(xc,'rows'); % remove duplicate permutations
    Sys.A = Sys.A(xc);
  end
  nSites = size(Sys.A,1);
  if numel(Sys.g)~=nSites
    Sys.g = Sys.g(ones(1,nSites));
  end
else
  if isfield(Sys,'A'), error('Sys.A given, but Sys.Nucs is missing.'); end
  if isfield(Sys,'n'), error('Sys.n given, but Sys.Nucs is missing.'); end
  nNucGroups = 0;
  nSites = numel(Sys.g);
end

if (nSites==1), error('Only one site given, no chemical exchange!'); end

% line width
if ~isfield(Sys,'lwpp'), Sys.lwpp = []; end
if ~isfield(Sys,'lw'), Sys.lw = []; end
if isempty(Sys.lw) && isempty(Sys.lwpp), error('Line width (Sys.lwpp or Sys.lw) is missing.'); end

if ~isempty(Sys.lw) && isempty(Sys.lwpp), Sys.lwpp = Sys.lw/sqrt(3); end

if (Sys.lwpp<=0), error('Linewidth (Sys.lw, Sys.lwpp) must be positive.'); end
if (numel(Sys.lwpp)~=1) && (numel(Sys.lwpp)~=nSites)
  error('Either 1 or %d line widths (Sys.lw, Sys.lwpp) expected for %d sites.',nSites,nSites);
end
if isfield(Sys,'Pop')&&(numel(Sys.Pop))~=nSites
    error('%d populations expected for %d sites.',nSites,nSites);
end
lwpp = mt2mhz(Sys.lwpp,mean(Sys.g)); % field to frequency conversion
Gamma = lwpp*pi*sqrt(3); % peak-to-peak lw -> Gamma (= 1/T2) conversion

% Exchange matrix and rates
%----------------------------------------------------------------
if ~isfield(Sys,'k'), Sys.k = []; end
if isempty(Sys.k), error('Sys.k (exchange rates/matrix) is missing.'); end
k = Sys.k;
twoSiteMutualExchange = (nSites==2) && (numel(k)==1);

if min(size(k))==1
  % mutual exchange: k contains linear list of rate constants
  nRates = nSites*(nSites-1)/2;
  if (numel(k)~=nRates)
    error('For %d-site mutual exchange, %d rate constants expected in Sys.k.',nSites,nRates);
  end
  if any(Sys.k<0), error('Sys.k cannot be negative.'); end
  X = zeros(nSites);
  X(tril(ones(nSites),-1)>0) = k;
  X = X + X.';
  X(1:nSites+1:end) = -sum(X);
else
  % general 2D exchange: k = general matrix for an arbitrary number of sites
  if (size(k,1)~=nSites) || (size(k,2)~=nSites) || numel(k)~=nSites^2
    error('2D exchange matrix in Sys.k has wrong size.');
  end
  X = k;
  if ~isfield(Sys,'Pop') %calculation of sites populations as formation-consumption ratio
      k(1:nSites+1:end) = 0; %force diagonal to zeroes
      P2 = sum(k,2); %formation rate as a sum of the row
      P1 = sum(k,1); %consumption rate as a sum of the column
      P1 = P1(:); %some tricks to form right dimension of the Sys.Pop
      Sys.Pop = sum(P2/P1,2);
  end
  X(1:nSites+1:end) = X(1:nSites+1:end) - sum(X,1); % set diagonal so column sums are 0
end

% Experimental parameters
%------------------------------------------------------------
if ~isfield(Exp,'mwFreq'), Exp.mwFreq = 9.5; end
if ~isfield(Exp,'nPoints'), Exp.nPoints = 2e3; end
if ~isfield(Exp,'Harmonic'), Exp.Harmonic = 1; end
if ~isfield(Exp,'Range'), Exp.Range = []; end
if ~isfield(Exp,'ModAmp'), Exp.ModAmp = 0; end
if isfield(Exp,'Temperature'), error('Exp.Temperature not supported.'); end

if ~((Exp.Harmonic==0) || (Exp.Harmonic==1) || (Exp.Harmonic==2))
  error('Exp.Harmonic must be 0, 1 or 2.');
end

AutoRange = isempty(Exp.Range);
if ~AutoRange
  if (numel(Exp.Range)~=2) || (Exp.Range(1)>=Exp.Range(2))
    error('Exp.Range is invalid. It must be of the form [Bmin Bmax].');
  end
end

% Options
%-------------------------------------------------------------
if ~isfield(Opt,'AutoRangeZoom'), Opt.AutoRangeZoom = 1.5; end
if ~isfield(Opt,'reduce'), Opt.reduce = 1; end
if numel(Gamma)~=1 Opt.reduce = 0; end %force disable reducing for different lws works properly
if ~isfield(Opt,'Output'), Opt.Output = 'summed'; end

separateOutput = strcmp(Opt.Output,'separate');

if ~isfield(Opt,'Method')
  if twoSiteMutualExchange
    if separateOutput, Opt.Method = 2; else Opt.Method = -3; end
  else
    if separateOutput, Opt.Method = 2; else Opt.Method = 1; end
  end
end

% Spectral shifts, positions and intensities
%-------------------------------------------------------------

% Shifts in the limit of no exchange
for iGroup = 1:nNucGroups
  % first-order perturbation limit
  if I(iGroup)~=0
      Int{iGroup} = equivsplit(I(iGroup),Sys.n(iGroup));
  else
      Int{iGroup} = 1;
  end
  mFmax = Sys.n(iGroup)*I(iGroup);
  mF = -mFmax:+mFmax;
  for iSite = 1:nSites
    shift{iGroup,iSite} = mF*Sys.A(iSite,iGroup);
  end
end

% choose a static field
% (sim is done as a freq sweep at this static field)
if AutoRange
  B0 = mhz2mt(Exp.mwFreq*1e3,mean(Sys.g));
else
  B0 = mean(Exp.Range); % center of given sweep range
end
nu0 = mt2mhz(B0,Sys.g)-Exp.mwFreq*1e3; % MHz

% Positions and intensities in the static limit of no exchange
%--------------------------------------------------------------
if (nNucGroups>0)
  for iSite = 1:nSites
    cumshift =  sum(allcombinations(shift{:,iSite}),2);
    nu(:,iSite) = nu0(iSite) + cumshift;
  end
  intensity = prod(allcombinations(Int{:}),2);
else
  nu = nu0;
  intensity = 1;
end
intensity = intensity/sum(intensity);
maxshift = max(abs(nu(:)));
nLines = numel(intensity);

% Reduction
%-----------------------------------------------------------------
% For two-site mutual exchange, almost half of the line pairs
% occur twice (e.g. nu(k1,:) = [-10 7] and nu(k2,:) = [7 -10])
% Line pairs occure once only if for each group of nuclei 
% mF of site 1 is equal to mF of site 2 (that is, the line
% position does not change upon exchange).
if twoSiteMutualExchange
  
  switch Opt.reduce
    case 0 % keep duplicates
      
    case 1 % remove duplicates
      %[nu intensity]
      nu = sort(nu,2);
      [nu,idx] = sortrows(nu);
      intensity = intensity(idx);
      
      keep = [true; any(abs(nu(1:end-1,:)-nu(2:end,:))>1e-5,2)];
      nu = nu(keep,:);
      fprintf('reduction: %d -> %d (%0.3g%%)\n',nLines,size(nu,1),100*size(nu,1)/nLines);
      
      iLine0 = 1;
      for iLine = 1:nLines
        if keep(iLine)
          iLine0 = iLine;
        else
          intensity(iLine0) = intensity(iLine0)+intensity(iLine);
        end
      end
      intensity = intensity(keep);
      nLines = size(nu,1);
      %[nu intensity]
      
    otherwise
      error('Unknown value for Opt.reduce')
  end
end

% Tresholding: remove lines with very low intensity
keep = intensity>0.001*max(intensity);
nu = nu(keep,:);
intensity = intensity(keep);
fprintf('thresholding: %d -> %d (%0.3g%%)\n',nLines,size(nu,1),100*size(nu,1)/nLines);
nLines = numel(intensity);

if AutoRange
  freqSweepRange = max(max(Gamma),maxshift)*Opt.AutoRangeZoom*2;
  Exp.Range = B0 + [-1 1]*mhz2mt(freqSweepRange/2,mean(Sys.g));
end
freqSweepRange = mt2mhz(diff(Exp.Range),mean(Sys.g));
nux = linspace(1,-1,Exp.nPoints)*freqSweepRange/2;

% Compute spectrum
%--------------------------------------------------------------
if separateOutput, nSpectra = nLines; else nSpectra = 1; end
spc = zeros(nSpectra,Exp.nPoints);

% Convert all frequenies to angular frequencies
nux = 2*pi*nux;
nu = 2*pi*nu;

switch (Opt.Method)
  case -3
    % Explicit method, two-site mutual exchange, Matlab accumulation
    %---------------------------------------------------------------
    dnu = nu(:,1) - nu(:,2);
    delta2 = (dnu/2).^2;
    numean = mean(nu,2);
    
    % Mex accumulation
    spc = exlinesum2(k,Gamma,numean,delta2,intensity,nux);
    
  case -2
    % Explicit method, two-site mutual exchange, mex accumulation
    %-------------------------------------------------------------
    dnu = nu(:,1) - nu(:,2);
    delta2 = (dnu/2).^2;
    numean = mean(nu,2);
    
    % Matlab accumulation
    a = 2*k+Gamma;
    c0 = Gamma*(2*k+Gamma);
    d0 = 2*(k+Gamma);
    for iLine = 1:nLines
      xi = nux - numean(iLine);
      b = xi;
      c = c0 + delta2(iLine) - xi.^2;
      d = d0*xi;
      spc_ = intensity(iLine)*(a*c+b.*d)./(c.^2+d.^2);
      if separateOutput
        spc(iLine,:) = spc_;
      else
        spc = spc + spc_;
      end
    end

  case -1
    % Liouville method, two-site mutual exchange, Matlab accumulation
    %----------------------------------------------------------------
    
    if ~twoSiteMutualExchange
      error('This method works only for two-site mutual exchange.');
    end
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
      
    % S and L computation
    dnu = nu(:,1)-nu(:,2);
    q = sqrt(4*k^2-dnu.^2)/2;
    numean = mean(nu,2);
    Svalues = repmat(intensity,1,nSites).*[1-k./q, 1+k./q]/2;
    Lvalues = [-Gamma-k-q-1i*numean, -Gamma-k+q-1i*numean];
        
    % accumulation
    % treat special case 4k^2 == dnu using one-line expression
    idx = find(q==0);
    for ii = 1:numel(idx)
      zeta = Gamma+k+1i*(nux-numean(idx(ii)));
      spc = spc + intensity(idx(ii))*real((k+zeta)./zeta.^2);
    end

    % accumulation
    for iLine = 1:numel(Lvalues)
      S = Svalues(iLine);
      if isinf(S) || isnan(S), continue; end
      L = Lvalues(iLine);
      x = imag(L)+nux;
      spc_ = -(real(L)*real(S)+imag(S)*x)./(real(L)^2+x.^2);
      spc = spc + spc_;
    end
    
  case 0
    
    % Liouville method, two-site mutual exchange, mex accumulation
    %-------------------------------------------------------------
    
    if ~twoSiteMutualExchange
      error('This method works only for two-site mutual exchange.');
    end
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
    
    % S and L computation
    dnu = abs(nu(:,1)-nu(:,2));
    q = sqrt(4*k^2-dnu.^2)/2;
    numean = mean(nu,2);
    Svalues = repmat(intensity,1,nSites).*[1-k./q, 1+k./q]/2;
    Lvalues = [-Gamma-k-q-1i*numean, -Gamma-k+q-1i*numean];
    
    % accumulation
    % (a) treat special case 4k^2==dnu using one-term expression
    %     (Svalues(:,iLine)==inf in this case)
    spc = zeros(1,Exp.nPoints);
    idx = find(q==0);
    for iLine = 1:numel(idx)
      zeta = Gamma+k+1i*(nux-numean(idx(iLine)));
      spc = spc + intensity(idx(iLine))*real((k+zeta)./zeta.^2);
    end
    
    % (b) general case: mex accumulation
    %     (omits lines with Svalues(:,iLine)==inf)
    spc = spc + exlinesum(Svalues,Lvalues,nux);
        
  case 1
    
    % Liouville method, general n-site exchange, mex accumulation
    %---------------------------------------------------------------
    if separateOutput
      error('Opt.Output=''separate'' not supported with this method.');
    end
    
    % S and L computation
    if isfield(Sys,'Pop')
        Pop = Sys.Pop(:)/sum(Sys.Pop); % fractional populations
    else
        Pop = ones(nSites,1)/nSites; % fractional populations
    end
    Det = ones(1,nSites); % detection operator
    if numel(Gamma)==1
        M0 = -Gamma*eye(nSites) + X;
    else
        M0 = diag(-Gamma) + X;
    end
    Lvalues = zeros(nSites,nLines);
    Svalues = zeros(nSites,nLines);
    for iLine = 1:nLines
      [U,Lambda] = eig(M0 - diag(1i*nu(iLine,:)));
      Lvalues(:,iLine) = diag(Lambda);
      Svalues(:,iLine) = (Det*U).'.*(U\Pop);
    end
    Svalues = Svalues.*repmat(intensity.',nSites,1);
    
    % accumulation
    spc = exlinesum(Svalues,Lvalues,nux);
    
  case 2
    
    % Liouville method, general n-site exchange, Matlab accumulation
    %---------------------------------------------------------------
    
    % S and L computation
    if isfield(Sys,'Pop')
        Pop = Sys.Pop(:)/sum(Sys.Pop); % fractional populations
    else
        Pop = ones(nSites,1)/nSites; % fractional populations
    end
    Det = ones(1,nSites);% % detection operator
    if numel(Gamma)==1
        M0 = -Gamma*eye(nSites) + X;
    else
        M0 = diag(-Gamma) + X;
    end
    Lvalues = zeros(nSites,nLines);
    Svalues = zeros(nSites,nLines);
    for iLine = 1:nLines
      [U,Lambda] = eig(M0 - diag(1i*nu(iLine,:)));
      Lvalues(:,iLine) = diag(Lambda);
      Svalues(:,iLine) = (Det*U).'.*(U\Pop);
    end
    Svalues = Svalues.*repmat(intensity.',nSites,1);
    
    % accumulation
    for iLine=1:nLines
      spc_ = 0;
      for iSite=1:nSites
        S = Svalues(iSite,iLine);
        L = Lvalues(iSite,iLine);
        x = imag(L)+nux;
        spc_ = spc_ -(real(L)*real(S)+imag(S)*x)./(real(L)^2+x.^2);
      end
      if separateOutput
        spc(iLine,:) = spc_;
      else
        spc = spc + spc_;
      end
    end
    
  case 3
    
    % Bloch/McConnell, two-site mutual exchange
    %-------------------------------------------------------------
    % my own version
    % (for another form see also Weil/Bolton p.320, Eq.(10.28b))
    
    if ~twoSiteMutualExchange
      error('Bloch/McConnel expressions implemented for two-site mutual exchange only.')
    end
    dnu = nu(:,1)-nu(:,2);
    q2 = k^2 - dnu.^2/4;
    numean = mean(nu,2);
    zeta0 =  Gamma + k + 1i*nux;
    for iLine = 1:nLines
      zeta = zeta0 - 1i*numean(iLine);
      Z_ = (k+zeta)./(q2(iLine)-zeta.^2);
      Z_ = -intensity(iLine)*real(Z_);
      if separateOutput
        spc(iLine,:) = Z_;
      else
        spc = spc + Z_;
      end
    end

  otherwise
    error('Unknown method.');
end

B = linspace(Exp.Range(1),Exp.Range(2),Exp.nPoints);


% Field modulation
%-------------------------------------------------------------
if (Exp.Harmonic>0)
  if (Exp.ModAmp==0)
    for m=1:Exp.Harmonic, spc = deriv(B,spc.').'; end
  else
    spc = fieldmod(B,spc,Exp.ModAmp,Exp.Harmonic);
  end
end

% Gaussian broadening
%-------------------------------------------------------------
if isfield(Sys,'lwg')
    df=(max(Exp.Range)-min(Exp.Range))/Exp.nPoints; %field step for convspec
    spc=convspec(spc,df,[0 Sys.lwg],0,1);
end
% Output
%-------------------------------------------------------------
switch nargout
  case 0
    plot(B,spc);
    title(sprintf('%0.8g GHz',Exp.mwFreq));
    axis tight
    xlabel('magnetic field (mT)');
  case 1
    varargout = {spc};
  case 2
    varargout = {B,spc};
  otherwise
    error('Wrong number of output arguments.');
end

function Combs = allcombinations(varargin)

if (nargin==0), Combs = []; return; end

Combs = varargin{1}(:);
nCombs = numel(Combs);

for iArg = 2:nargin
  New = varargin{iArg}(:);  nNew = numel(New);
  [idxNew,idxCombs] = find(ones(nNew,nCombs));
  Combs = [Combs(idxCombs(:),:), New(idxNew(:))];
  nCombs = nCombs*nNew;
end

return
